#include "FdIO.h"

/*-----------------------------------------------declare private functions---*/
typedef struct FdRescale  FdRescale;

struct FdRescale{
  char*      output;      /* output channel name                             */
  double     scale;       /* scaling factor for input                        */
  char*      input;       /* input channel name                              */
  char*      units;       /* new units if the channel content has been changed
			     or NULL if the units are unchanged              */
  FRBOOL     present;     /* to control the "present" error message          */
  FRBOOL     alreadyThere;/* to control the "laready there" error message    */
  FdAction*  action;      /* action object which hold this recast object     */
};
 
int  FdRescaleNew(void *arg, CfgDataType_t *data);
void FdRescaleProcess(FdRescale* list, FrameH* frame);
 
/*---------------------------------------------------------------------------*/
int FdRescaleNew(void *arg, CfgDataType_t *data)
/*---------------------------------------------------------------------------*/
{
  FdRescale *r;

  /*------------create the object and add it at the end of the linked list---*/
  r = (FdRescale*) calloc(1,sizeof(FdRescale));
  if(r == NULL) CfgMsgAddFatal("malloc FdRescale failed");

  r->action = FdActionNew((FdAction**) arg, 
			       (void*) r, FdRescaleProcess, "Rescale");
  /*------------------------------------------------------------ fill info---*/
  FrStrCpy(&r->output, CfgParseGetNextString(data));
  r->scale =           CfgParseGetNextReal(data);
  FrStrCpy(&r->input,  CfgParseGetNextString(data));
  FrStrCpy(&r->units,  CfgParseGetNextString(data));
  r->present = FR_YES;
  r->alreadyThere = FR_NO;

  CfgMsgAddInfo("New output channel %s = %g*%s units=%s",
		r->output, r->scale, r->input, r->units);
 
  return(CFG_OK);
}
/*---------------------------------------------------------------------------*/
void FdRescaleProcessOne(FdRescale* r,
			FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FrAdcData *adc;
  FrProcData *proc;
  FRBOOL present, alreadyThere;

  present = FR_YES;

  /*-----------------first deal with the case were the channel is unchanged---*/
  if(strcmp(r->input, r->output) == 0) {

    if((adc = FrAdcDataFind(frame, r->input)) != NULL) {
      adc->slope *= r->scale;
      if(r->units != NULL) {
        free(adc->units);
        FrStrCpy(&(adc->units),r->units);}}
    
    else if((proc = FrProcDataFind(frame, r->input)) != NULL) {
      FrvScale(r->scale, proc->data, proc->data, NULL);
      if(r->units != NULL) FrVectSetUnitY(proc->data, r->units);}      

    else {present = FR_NO;}}

  /*-------create a new channel; first check that it is not already there---*/
  else {
    alreadyThere = FR_NO;
    if((FrAdcDataFind(frame, r->output) != NULL) ||
       (FrProcDataFind(frame, r->output) != NULL)){
      alreadyThere = FR_YES;}

    else if((adc = FrAdcDataFind(frame, r->input)) != NULL) {
      adc = FrAdcDataCopy(adc, frame);
      if(adc == NULL) CfgMsgAddFatal("FdRescaleProcessOne: malloc failed");
      free(adc->name);
      FrStrCpy(&(adc->name),r->output);
      FrVectSetName(adc->data, r->output);
      adc->slope *= r->scale;
      if(r->units != NULL) {
        free(adc->units);
        FrStrCpy(&(adc->units),r->units);}}

    else if((proc = FrProcDataFind(frame, r->input)) != NULL) {
      proc = FrProcDataCopy(proc, frame);
      if(proc == NULL) CfgMsgAddFatal("FdRescaleProcessOne: malloc failed");
      free(proc->name);
      FrStrCpy(&(proc->name),r->output);
      FrVectSetName(proc->data, r->output);
      FrvScale(r->scale, proc->data, proc->data, NULL);
      if(r->units != NULL) FrVectSetUnitY(proc->data, r->units);}
 
    else {present = FR_NO;}

    if(alreadyThere != r->alreadyThere) {
      if(alreadyThere == FR_YES) 
        CfgMsgAddError("%d: channel %s is already there, could not rescale it",
                        frame->GTimeS, r->output);
      else
        CfgMsgAddInfo("%d: channel %s is no more already there, back for rescaling",
                    frame->GTimeS, r->output);
      r->alreadyThere = alreadyThere;}}

  /*-------------------------------------------------------error messages---*/
  if(present != r->present) {
    if(present == FR_NO) 
      CfgMsgAddError("%d: channel %s is missing, could not rescale it",
                      frame->GTimeS, r->input);
    else
      CfgMsgAddInfo("%d: channel %s is back for rescaling",
                    frame->GTimeS, r->input);
    r->present = present;}

  return;
}
/*---------------------------------------------------------------------------*/
void FdRescaleProcess(FdRescale *combine,
		     FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FdAction *next;

  if(frame != NULL) FdRescaleProcessOne(combine, frame);

  next = combine->action->next;
  if(next != NULL) next->action(next->data, frame);

  return;
}
/*---------------------------------------------------------------------------*/
void FdRescaleParserAdd(FdIO* fdIO)
/*---------------------------------------------------------------------------*/
{
 
  CfgParseGetFunctionAdd(fdIO->parser, "FDIN_RESCALE_CHANNEL",
			 FdRescaleNew, (void *) &(fdIO->actionsIn), 4,
			 CfgString, CfgReal, CfgString, CfgString);
  CfgParseGetFunctionAdd(fdIO->parser, "FDOUT_RESCALE_CHANNEL",
			 FdRescaleNew, (void *) &(fdIO->actionsOut), 4,
			 CfgString, CfgReal, CfgString, CfgString);

  return;
}
